from __future__ import annotations

import re
from pathlib import Path

import jproperties  # pyright: ignore[reportMissingTypeStubs]
from packaging.version import Version
from pydantic import BaseModel

GRADLE_VERSION_RE = re.compile(
    r"""
    v?
    (?P<version>
        \d+
        (?:\.\d+)*
    )
    (?:
        -
        (?P<rc>\d+)
    )?
    """,
    re.VERBOSE,
)


class GradleVersion(BaseModel):
    version: str
    rc: int | None

    @classmethod
    def from_properties(cls, p: jproperties.Properties, key: str):
        raw_version = str(p[key].data)

        match = GRADLE_VERSION_RE.match(raw_version)
        if match is None:
            raise ValueError(f"Failed to parse version {key}={raw_version}")

        return cls.model_validate(match.groupdict())

    def full_version(
        self,
        py_version: str | Version,
        *,
        next_rc: bool = False,
    ) -> str:
        # ensure py_version is a valid version by itself
        if isinstance(py_version, str):
            py_version = Version(py_version)

        if py_version.pre:
            raise ValueError("a/b/c/pre/rc is reserved for Gradle prereleases")

        # split py_version at the point where we need to insert the gradle prerelease
        py_base = py_version.base_version
        py_rest = str(py_version).removeprefix(py_base)

        # construct the full version
        # eg. 1.2.3 . 4.5 rc6 .dev7
        rc = self.rc_segment(next_rc)
        full_version = self.version + "." + py_base + rc + py_rest

        # round-trip through Version to normalize it
        return str(Version(full_version))

    def rc_segment(self, next_rc: bool):
        if self.rc is None:
            return ""
        return f"rc{self.next_rc if next_rc else self.rc}"

    @property
    def next_rc(self):
        if self.rc is None:
            raise ValueError("Tried to call next_rc on a non-rc version")
        return self.rc + 1


def load_properties(path: Path):
    p = jproperties.Properties()
    with open(path, "rb") as f:
        p.load(f, "utf-8")
    return p



from functools import cached_property
from pathlib import Path
from typing import Any

import jproperties  # pyright: ignore[reportMissingTypeStubs]
from hatchling.bridge.app import Application
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from pydantic import BaseModel, Field

from ..common.gradle import GradleVersion, load_properties


class GradleDependency(BaseModel):
    package: str
    op: str
    key: str
    py_version: str = Field(alias="py-version")
    rc_upper_bound: bool = Field(alias="rc-upper-bound", default=False)
    """If True and gradle_version has a pre-release suffix (eg. `0.1.0-1`), add a
    corresponding exclusive upper bound for the next RC version (eg. `<0.1.0.1.0rc2`).
    
    If True, the operators `>=` and `~=` effectively become `==`. If False, pip may
    install a later prerelease or a released version. There's not really a "best"
    option, which is why this flag exists.

    The default is `False`. We think this *should* work in more cases than not.
    """

    def version_specifier(self, p: jproperties.Properties, app: Application):
        gradle = self.gradle_version(p)

        full_version = gradle.full_version(self.py_version)
        lower_bound = self.op + full_version

        if not (gradle.rc is not None and self.rc_upper_bound):
            return self.package + lower_bound

        if "<" not in self.op:
            app.display_warning(
                f"WARNING: Dependency on package {self.package} will ONLY accept {full_version} (because gradle_version {self.gradle_version} is a prerelease)."
            )

        upper_bound = "<" + gradle.full_version(self.py_version, next_rc=True)
        return f"{self.package}{lower_bound},{upper_bound}"

    def gradle_version(self, p: jproperties.Properties):
        return GradleVersion.from_properties(p, self.key)


class GradlePropertiesBuildHookConfig(BaseModel):
    path: Path = Path("gradle.properties")
    gradle_dependencies: list[GradleDependency] = Field(alias="gradle-dependencies")


class GradlePropertiesBuildHook(BuildHookInterface[Any]):
    PLUGIN_NAME = "gradle-properties"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        build_data.setdefault("dependencies", [])

        p = load_properties(self.path)
        for dependency in self.gradle_dependencies:
            build_data["dependencies"].append(dependency.version_specifier(p, self.app))

    @cached_property
    def typed_config(self):
        return GradlePropertiesBuildHookConfig.model_validate(self.config)

    @property
    def path(self):
        return self.typed_config.path

    @property
    def gradle_dependencies(self):
        return self.typed_config.gradle_dependencies



from functools import cached_property
from pathlib import Path
from typing import Any

from hatchling.version.source.regex import RegexSource
from pydantic import BaseModel, Field

from ..common.gradle import GradleVersion, load_properties


class PropertiesVersionSourceConfig(BaseModel):
    py_path: str = Field(alias="py-path")

    gradle_path: Path = Field(alias="gradle-path", default=Path("gradle.properties"))
    key: str = "modVersion"


class PropertiesVersionSource(RegexSource):
    PLUGIN_NAME = "gradle-properties"

    def get_version_data(self) -> dict[str, Any]:
        # load gradle version from gradle.properties
        p = load_properties(self.gradle_path)
        gradle = GradleVersion.from_properties(p, self.key)

        # use the regex source to load py_version from a file
        self.config["path"] = self.py_path
        py_version_data = super().get_version_data()
        py_version = py_version_data["version"]

        return py_version_data | {
            "version": gradle.full_version(py_version),
            "gradle_version": gradle,
            "py_version": py_version,
        }

    def set_version(self, version: str, version_data: dict[str, Any]) -> None:
        super().set_version(version_data["py_version"], version_data)

    # config values

    @cached_property
    def typed_config(self):
        return PropertiesVersionSourceConfig.model_validate(self.config)

    @property
    def gradle_path(self):
        path = Path(self.root) / self.typed_config.gradle_path
        if not path.is_file():
            raise FileNotFoundError(
                f"File does not exist or is not a file: {self.typed_config.gradle_path}"
            )
        return path

    @property
    def key(self):
        return self.typed_config.key

    @property
    def py_path(self):
        return self.typed_config.py_path



from hatchling.plugin import hookimpl

from .build_hook import GradlePropertiesBuildHook
from .version_scheme import GradleVersionScheme
from .version_source import PropertiesVersionSource


@hookimpl
def hatch_register_version_source():
    return PropertiesVersionSource


@hookimpl
def hatch_register_version_scheme():
    return GradleVersionScheme


@hookimpl
def hatch_register_build_hook():
    return GradlePropertiesBuildHook



from pathlib import Path
from typing import Any

import pytest

from ..common.cd import cd
from .build_hook import GradlePropertiesBuildHook


@pytest.mark.parametrize(
    # fmt: off
    "package, op,key,py_version,gradle_version,rc_upper_bound,full_version",
    [
        ("P", "~=", "KEY", "4",       "1.2.3",   False, "P~=1.2.3.4"),
        ("P", "~=", "KEY", "4.5",     "1.2.3",   False, "P~=1.2.3.4.5"),
        ("P", ">=", "KEY", "4.5",     "1.2.3",   False, "P>=1.2.3.4.5"),
        ("P", "~=", "KEY", "4.5",     "1.2.3-6", False, "P~=1.2.3.4.5rc6"),
        ("P", "~=", "KEY", "4.5",     "1.2.3-6", True,  "P~=1.2.3.4.5rc6,<1.2.3.4.5rc7"),
        ("P", "~=", "KEY", "4.5dev6", "1.2.3",   False, "P~=1.2.3.4.5.dev6"),
        ("P", "~=", "KEY", "4.5dev7", "1.2.3-6", False, "P~=1.2.3.4.5rc6.dev7"),
        ("P", "~=", "KEY", "4.5dev8", "1.2.3-6", True,  "P~=1.2.3.4.5rc6.dev8,<1.2.3.4.5rc7.dev8"),
    ],
    # fmt: on
)
def test_gradle_properties_deps(
    tmp_path: Path,
    package: str,
    op: str,
    key: str,
    py_version: str,
    gradle_version: str,
    rc_upper_bound: bool,
    full_version: str,
):
    # arrange
    hook = GradlePropertiesBuildHook(
        root="",
        config={
            "gradle-dependencies": [
                {
                    "package": package,
                    "op": op,
                    "key": key,
                    "py-version": py_version,
                    "rc-upper-bound": rc_upper_bound,
                }
            ],
        },
        build_config=None,
        metadata=None,  # type: ignore
        directory="",
        target_name="",
        app=None,
    )
    build_data = dict[str, Any]()
    (tmp_path / "gradle.properties").write_text(f"{key}={gradle_version}")

    # act
    with cd(tmp_path):
        hook.initialize("", build_data)

    # assert
    assert build_data["dependencies"] == [full_version]
